# -*- coding: utf-8 -*-

from plone.jsonapi.routes import add_plone_route

# CRUD
from plone.jsonapi.routes.api import get_record
from plone.jsonapi.routes.api import create_items
from plone.jsonapi.routes.api import update_items
from plone.jsonapi.routes.api import delete_items

from plone.jsonapi.routes.api import url_for


# GET
@add_plone_route("/get", "get", methods=["GET"])
@add_plone_route("/get/<string:uid>", "get", methods=["GET"])
def get(context, request, uid=None):
    """ get content
    """
    return get_record(uid=uid)


# CREATE
@add_plone_route("/create", "create", methods=["POST"])
@add_plone_route("/create/<string:uid>", "create", methods=["POST"])
def create(context, request, uid=None):
    """ create content
    """
    items = create_items(uid=uid)
    return {
        "url": url_for("create"),
        "count": len(items),
        "items": items,
    }


# UPDATE
@add_plone_route("/update", "update", methods=["POST"])
@add_plone_route("/update/<string:uid>", "update", methods=["POST"])
def update(context, request, uid=None):
    """ update content
    """
    items = update_items(uid=uid)
    return {
        "url": url_for("update"),
        "count": len(items),
        "items": items,
    }


# DELETE
@add_plone_route("/delete", "delete", methods=["POST"])
@add_plone_route("/delete/<string:uid>", "delete", methods=["POST"])
def delete(context, request, uid=None):
    """ delete content
    """
    items = delete_items(uid=uid)
    return {
        "url": url_for("delete"),
        "count": len(items),
        "items": items,
    }

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

__author__    = 'Ramon Bartl <ramon.bartl@googlemail.com>'
__docformat__ = 'plaintext'

import logging

from zope import interface

from plone import api

from AccessControl import Unauthorized

from plone.jsonapi.routes.interfaces import IDataManager

logger = logging.getLogger("plone.jsonapi.routes.datamanagers")


class ATDataManager(object):
    """ Adapter to set and get field values of AT Content Types
    """
    interface.implements(IDataManager)

    def __init__(self, context):
        self.context = context

    def is_file_field(self, field):
        """ checks if field is a file field
        """
        # XXX find a better way to distinguish file/image fields
        if field.__name__ in ["file", "image"]:
            return True
        return False

    def get_field(self, name):
        return self.context.getField(name)

    def set(self, name, value):
        field = self.get_field(name)
        logger.info("ATDataManager::set: name=%r, value=%r, field=%r", name, value, field)
        if not field:
            return False
        if not field.checkPermission("write", self.context):
            raise Unauthorized("You are not allowed to write the field %s" % name)
        if self.is_file_field(field):
            logger.info("ATDataManager::set:File field detected ('%r'), base64 decoding value", field)
            value = str(value).decode("base64")
        mutator = field.getMutator(self.context)
        mutator(value)
        return True

    def get(self, name):
        field = self.get_field(name)
        if not field.checkPermission("read", self.context):
            raise Unauthorized("You are not allowed to read the field %s" % name)
        return field.get(self.context)


class DexterityDataManager(object):
    """ Adapter to set and get field values of Dexterity Content Types
    """
    interface.implements(IDataManager)

    def __init__(self, context):
        self.context = context
        self.schema = self.get_schema()

    def get_schema(self):
        pt = api.portal.get_tool("portal_types")
        fti = pt.getTypeInfo(self.context.portal_type)
        return fti.lookupSchema()

    def get_field(self, name):
        return self.schema.get(name)

    def set(self, name, value):
        field = self.get_field(name)
        logger.info("DexterityDataManager::set: name=%r, value=%r, field=%r", name, value, field)
        field.set(self.context, value)

    def get(self, name):
        field = self.get_field(name)
        return field.get(self.context)

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

from plone.jsonapi.routes import add_plone_route

from plone.jsonapi.routes.api import get_record


# GET
@add_plone_route("/portal", "portal", methods=["GET"])
@add_plone_route("/portal/<string:uid>", "portal", methods=["GET"])
def get(context, request, uid=0):
    """ get the Portal
    """
    return get_record(uid=uid)

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

__author__    = 'Ramon Bartl <ramon.bartl@nexiles.com>'
__docformat__ = 'plaintext'

import pkg_resources
from plone.jsonapi.routes import api
from plone.jsonapi.routes import add_plone_route

def version():
    dist = pkg_resources.get_distribution("plone.jsonapi.routes")
    return dist.version

__version__ = version()
__build__ = 360
__date__ = '2015-02-22'


@add_plone_route("/version", "ploneapiversion", methods=["GET"])
def apiversion(context, request):
    """ get the current version of this package
    """
    return {
        "url":     api.url_for("ploneapiversion"),
        "version": __version__,
        "build":   __build__,
        "date":    __date__,
    }

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

from plone.jsonapi.routes import add_plone_route

# CRUD
from plone.jsonapi.routes.api import get_batched
from plone.jsonapi.routes.api import create_items
from plone.jsonapi.routes.api import update_items
from plone.jsonapi.routes.api import delete_items

from plone.jsonapi.routes.api import url_for


# GET
@add_plone_route("/files", "files", methods=["GET"])
@add_plone_route("/files/<string:uid>", "files", methods=["GET"])
def get(context, request, uid=None):
    """ get files
    """
    return get_batched("File", request, uid=uid, endpoint="files")


# CREATE
@add_plone_route("/files/create", "files_create", methods=["POST"])
@add_plone_route("/files/create/<string:uid>", "files_create", methods=["POST"])
def create(context, request, uid=None):
    """ create files
    """
    items = create_items("File", request, uid=uid, endpoint="files")
    return {
        "url": url_for("files_create"),
        "count": len(items),
        "items": items,
    }


# UPDATE
@add_plone_route("/files/update", "files_update", methods=["POST"])
@add_plone_route("/files/update/<string:uid>", "files_update", methods=["POST"])
def update(context, request, uid=None):
    """ update files
    """
    items = update_items("File", request, uid=uid, endpoint="files")
    return {
        "url": url_for("files_update"),
        "count": len(items),
        "items": items,
    }


# DELETE
@add_plone_route("/files/delete", "files_delete", methods=["POST"])
@add_plone_route("/files/delete/<string:uid>", "files_delete", methods=["POST"])
def delete(context, request, uid=None):
    """ delete files
    """
    items = delete_items("File", request, uid=uid, endpoint="files")
    return {
        "url": url_for("files_delete"),
        "count": len(items),
        "items": items,
    }

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

import doctest

import unittest2 as unittest

from Testing import ZopeTestCase as ztc

from plone.jsonapi.routes.tests.base import APITestCase

def test_suite():
    suite = unittest.TestSuite()
    suite.addTests([
        ztc.ZopeDocFileSuite(
            '../docs/Readme.txt',
            test_class=APITestCase,
            optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE,
        ),
    ])
    return suite

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

import json

from plone.app.testing import TEST_USER_ID
from plone.app.testing import setRoles

from plone.jsonapi.routes.tests.base import APITestCase
from plone.jsonapi.routes import api



class TestAPI(APITestCase):
    """ Test the API functionality
    """

    def setUp(self):
        self.portal  = self.getPortal()
        self.request = self.getRequest()

        # the test folder we created
        self.test_folder = self.portal.folder

        # give the test user the manager role
        setRoles(self.portal, TEST_USER_ID, ['Manager'])

    def get_document_obj(self):
        _, obj = self.test_folder.items()[0]
        return obj

    def get_document_brain(self):
        pc = api.get_portal_catalog()
        results = pc({"portal_type": "Document", "limit":1})
        return results[0]

    def test_get_portal(self):
        self.assertEqual(api.get_portal(), self.portal)

    def test_is_brain(self):
        brain = self.get_document_brain()
        self.assertTrue(api.is_brain(brain))
        self.assertFalse(api.is_brain(api.get_object(brain)))

    def test_get_uid(self):
        obj = self.get_document_obj()
        self.assertEqual(api.get_uid(obj), obj.UID())

    def test_get_parent(self):
        parent = self.test_folder
        obj = self.get_document_obj()
        self.assertEqual(api.get_parent(obj), parent)

    def test_get_object(self):
        obj = self.get_document_obj()
        brain = self.get_document_brain()
        self.assertEqual(api.get_object(obj), obj)
        self.assertEqual(api.get_object(brain), brain.getObject())

    def test_get_object_by_uid(self):
        obj = self.get_document_obj()
        uid = obj.UID()
        self.assertEqual(api.get_object_by_uid(uid), obj)
        # specail case: UID 0 is the portal
        self.assertEqual(api.get_object_by_uid(0), self.portal)


class TestCRUDAPI(APITestCase):
    """ Test the CREATE/UPDATE/DELETE functionality of the API
    """

    def setUp(self):
        self.portal  = self.getPortal()
        self.request = self.getRequest()

        # the test folder we created
        self.test_folder = self.portal.folder

        # give the test user the manager role
        setRoles(self.portal, TEST_USER_ID, ['Manager'])

    def test_get_items(self):
        results = api.get_items("Folder")
        self.assertEqual(len(results), 1)

    def test_get_batched(self):
        results = api.get_batched("Document")
        self.assertEqual(results["count"], 50)
        self.assertEqual(results["pages"], 2)

    def test_create_items(self):
        # create a folder inside the testfolder
        data = {"title": "Subfolder"}
        self.request["BODY"] = json.dumps(data)
        api.create_items("Folder", uid=self.test_folder.UID())
        results = api.get_items("Folder")
        self.assertEqual(len(results), 2)
        self.assertEqual(self.test_folder.subfolder.Title(), "Subfolder")

    def test_update_items(self):
        data = {"title": "Test Folder Changed"}
        self.request["BODY"] = json.dumps(data)
        self.assertEqual(self.test_folder.Title(), "Test Folder")
        api.update_items("Folder", uid=self.test_folder.UID())
        self.assertEqual(self.test_folder.Title(), data["title"])

    def test_delete_items(self):
        api.delete_items("Folder", uid=self.test_folder.UID())
        self.assertEqual(len(api.get_items("Folder")), 0)


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestAPI))
    suite.addTest(makeSuite(TestCRUDAPI))
    return suite

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

import types
import logging

from plone.jsonapi.routes.api import url_for

__author__ = 'Ramon Bartl <ramon.bartl@googlemail.com>'
__docformat__ = 'plaintext'

logger = logging.getLogger("plone.jsonapi.routes")


def returns_plone_items_for(endpoint):
    """ returns a dictionary with items
    """

    def decorator(func):
        def inner(*args, **kwargs):
            result = func(*args, **kwargs)
            if type(result) is types.ListType:
                return {
                    "url": url_for(endpoint),
                    "count": len(result),
                    "items": result,
                }
            return result
        return inner
    return decorator



# -*- coding: utf-8 -*-

import json
import logging

from zope.globalrequest import getRequest

from plone.jsonapi.routes import underscore as _

__author__ = 'Ramon Bartl <ramon.bartl@googlemail.com>'
__docformat__ = 'plaintext'

logger = logging.getLogger("plone.jsonapi.routes.request")


def get_request():
    """ return the request object
    """
    return getRequest()


def get_form():
    """ return the request form dictionary
    """
    return get_request().form


def get(key, default=None):
    """ return the key from the request
    """
    return get_form().get(key, default)


def get_complete():
    """ returns the 'complete' from the request
    """
    complete = get("complete", "no")
    if complete.lower() in ["y", "yes", "1", "true"]:
        return True
    return False


def get_children():
    """ returns the 'children' from the request
    """
    complete = get("children", "no")
    if complete.lower() in ["y", "yes", "1", "true"]:
        return True
    return False


def get_sort_limit():
    """ returns the 'sort_limit' from the request
    """
    limit = _.convert(get("sort_limit"), _.to_int)
    if (limit < 1):
        limit = None  # catalog raises IndexError if limit < 1
    return limit


def get_batch_size():
    """ returns the 'limit' from the request
    """
    return _.convert(get("limit"), _.to_int) or 25


def get_batch_start():
    """ returns the 'start' from the request
    """
    return _.convert(get("b_start"), _.to_int) or 0


def get_sort_on(allowed_indexes=None):
    """ returns the 'sort_on' from the request
    """
    sort_on = get("sort_on")
    if allowed_indexes and sort_on not in allowed_indexes:
        logger.warn("Index '%s' is not in allowed_indexes" % sort_on)
        return "id"
    return sort_on


def get_sort_order():
    """ returns the 'sort_order' from the request
    """
    sort_order = get("sort_order")
    if sort_order in ["ASC", "ascending", "a", "asc", "up", "high"]:
        return "ascending"
    if sort_order in ["DESC", "descending", "d", "desc", "down", "low"]:
        return "descending"
    # https://github.com/collective/plone.jsonapi.routes/issues/31
    return "ascending"


def get_query():
    """ returns the 'query' from the request
    """
    q = get("q", "")

    qs = q.lstrip("*.!$%&/()=#-+:'`´^")
    if qs and not qs.endswith("*"):
        qs += "*"
    return qs


def get_path():
    """ returns the 'path' from the request
    """
    return get("path", "")


def get_depth():
    """ returns the 'depth' from the request
    """
    return _.convert(get("depth", 0), _.to_int)


def get_recent_created():
    """ returns the 'recent_created' from the request
    """
    return get("recent_created", None)


def get_recent_modified():
    """ returns the 'recent_modified' from the request
    """
    return get("recent_modified", None)


def get_request_data():
    """ extract and convert the json data from the request

    returns a list of dictionaries
    """
    request = get_request()
    return _.convert(json.loads(request.get("BODY", "{}")), _.to_list)

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

from plone.jsonapi.routes import add_plone_route

# CRUD
from plone.jsonapi.routes.api import get_batched
from plone.jsonapi.routes.api import create_items
from plone.jsonapi.routes.api import update_items
from plone.jsonapi.routes.api import delete_items

from plone.jsonapi.routes.api import url_for


# GET
@add_plone_route("/folders", "folders", methods=["GET"])
@add_plone_route("/folders/<string:uid>", "folders", methods=["GET"])
def get(context, request, uid=None):
    """ get folders
    """
    return get_batched("Folder", uid=uid, endpoint="folders")


# CREATE
@add_plone_route("/folders/create", "folders_create", methods=["POST"])
@add_plone_route("/folders/create/<string:uid>", "folders_create", methods=["POST"])
def create(context, request, uid=None):
    """ create folders
    """
    items = create_items("Folder", uid=uid, endpoint="folders")
    return {
        "url": url_for("folders_create"),
        "count": len(items),
        "items": items,
    }


# UPDATE
@add_plone_route("/folders/update", "folders_update", methods=["POST"])
@add_plone_route("/folders/update/<string:uid>", "folders_update", methods=["POST"])
def update(context, request, uid=None):
    """ update folders
    """
    items = update_items("Folder", uid=uid, endpoint="folders")
    return {
        "url": url_for("folders_update"),
        "count": len(items),
        "items": items,
    }


# DELETE
@add_plone_route("/folders/delete", "folders_delete", methods=["POST"])
@add_plone_route("/folders/delete/<string:uid>", "folders_delete", methods=["POST"])
def delete(context, request, uid=None):
    """ delete folders
    """
    items = delete_items("Folder", uid=uid, endpoint="folders")
    return {
        "url": url_for("folders_delete"),
        "count": len(items),
        "items": items,
    }

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

__author__    = 'Ramon Bartl <ramon.bartl@googlemail.com>'
__docformat__ = 'plaintext'

import logging
import datetime
import DateTime

import simplejson as json

from zope import interface
from zope import component

from plone import api

from plone.dexterity.schema import SCHEMA_CACHE
from plone.dexterity.interfaces import IDexterityContent

from Products.CMFCore.interfaces import ISiteRoot
from Products.ZCatalog.interfaces import ICatalogBrain
from Products.ATContentTypes.interfaces import IATContentType

from plone.jsonapi.routes.interfaces import IInfo


logger = logging.getLogger("plone.jsonapi.routes")


class Base(object):
    """ Base Adapter
    """
    interface.implements(IInfo)

    def __init__(self, context):
        self.context = context
        self.keys = []

        # additional attributes to extract besides the Schema keys
        self.attributes = {
            "id":          "getId",
            "uid":         "UID",
            "title":       "Title",
            "description": "Description",
            "created":     "created",
            "modified":    "modified",
            "effective":   "effective",
            "portal_type": "portal_type",
            "tags":        "Subject",
        }

    def to_dict(self):
        data = to_dict(self.context, keys=self.keys)
        for key, attr in self.attributes.iteritems():
            if data.get(key): continue # don't overwrite
            value = getattr(self.context, attr, None)
            if callable(value):
                value = value()
            data[key] = get_value(value)
        return data

    def __call__(self):
        return self.to_dict()


class ZCDataProvider(Base):
    """ Catalog Brain Adapter
    """
    interface.implements(IInfo)
    component.adapts(ICatalogBrain)

    def __init__(self, context):
        super(ZCDataProvider, self).__init__(context)

    def to_dict(self):
        brain = self.context
        return {
            "id":          brain.getId,
            "uid":         brain.UID,
            "title":       brain.Title,
            "description": brain.Description,
            "url":         brain.getURL(),
            "portal_type": brain.portal_type,
            "created":     brain.created.ISO8601(),
            "modified":    brain.modified.ISO8601(),
            "effective":   brain.effective.ISO8601(),
            "type":        brain.portal_type,
            "tags":        brain.Subject,
        }


class DexterityDataProvider(Base):
    """ Data Provider for Dexterity based content types
    """
    interface.implements(IInfo)
    component.adapts(IDexterityContent)

    def __init__(self, context):
        super(DexterityDataProvider, self).__init__(context)

        schema = SCHEMA_CACHE.get(context.portal_type)
        self.keys = schema.names()


class ATDataProvider(Base):
    """ Archetypes Adapter
    """
    interface.implements(IInfo)
    component.adapts(IATContentType)

    def __init__(self, context):
        super(ATDataProvider, self).__init__(context)
        schema = context.Schema()
        self.keys = schema.keys()


class SiteRootDataProvider(Base):
    """ Site Root Adapter
    """
    interface.implements(IInfo)
    component.adapts(ISiteRoot)

    def __init__(self, context):
        super(SiteRootDataProvider, self).__init__(context)

#---------------------------------------------------------------------------
#   Functional Helpers
#---------------------------------------------------------------------------

def to_dict(obj, keys):
    """ returns a dictionary of the given keys
    """
    out = dict()
    for key in keys:
        field = get_field(obj, key)
        out[key] = get_value(field)
    if out.get("workflow_info"):
        logger.warn("Workflow Info ommitted since the key 'workflow_info' was ",
                "found in the current schema")
        return out
    wf_info = get_wf_info(obj)
    out["workflow_info"] = wf_info
    out["state"] = wf_info.get("status")
    return out


def get_field(obj, key):
    """ get the value for the given key
        => handles Dexterity/AT Content types
    """
    if IATContentType.providedBy(obj):
        return obj.getField(key).getAccessor(obj)()
    return getattr(obj, key)


def get_value(field):
    """ extract the value from the given field
    """

    if isinstance(field, (datetime.datetime, datetime.date, DateTime.DateTime)):
        return get_iso_date(field)

    if hasattr(field, "filename"):
        return get_file_dict(field)

    if not is_json_serializable(field):
        return None

    return field


def get_file_dict(field):
    """ file representation of the given data
    """

    data = field.data.encode("base64")
    content_type = getattr(field, "content_type", "application/octet-stream")

    return {
        "data": data,
        "size": len(field.data),
        "content_type": content_type
    }


def get_iso_date(date=None):
    """ get the iso string for python datetime objects
    """
    if date is None:
        return ""

    if isinstance(date, (DateTime.DateTime)):
        return date.ISO8601()

    return date.isoformat()


def is_json_serializable(thing):
    """ checks if the given thing can be serialized to json
    """
    try:
        json.dumps(thing)
        return True
    except TypeError:
        return False


def get_wf_info(obj):
    """ returns the workflow information of the first assigned workflow
    """

    # get the portal workflow tool
    wf_tool = api.portal.get_tool("portal_workflow")

    # the assigned workflows of this object
    wfs = wf_tool.getWorkflowsFor(obj)

    # no worfkflows assigned -> return
    if not wfs: return {}

    # get the first one
    workflow = wfs[0]

    # get the status info of the current state (dictionary)
    status = wf_tool.getStatusOf(workflow.getId(), obj)

    # https://github.com/collective/plone.jsonapi.routes/issues/33
    if not status:
        return {}

    # get the current review_status
    current_state_id = status.get("review_state", None)

    # get the wf status object
    current_status = workflow.states[current_state_id]

    # get the title of the current status
    current_state_title = current_status.title

    # get the transition informations
    transitions = map(to_transition_info, wf_tool.getTransitionsFor(obj))

    return {
        "workflow":     workflow.getId(),
        "status":       current_state_title,
        "review_state": current_state_id,
        "transitions":  transitions
    }


def to_transition_info(transition):
    """ return the transition information
    """
    return {
        "value":   transition["id"],
        "display": transition["description"],
        "url":     transition["url"],
    }

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

from plone.jsonapi.routes import add_plone_route

# CRUD
from plone.jsonapi.routes.api import get_batched
from plone.jsonapi.routes.api import create_items
from plone.jsonapi.routes.api import update_items
from plone.jsonapi.routes.api import delete_items
from plone.jsonapi.routes.decorators import returns_plone_items_for

from plone.jsonapi.routes.api import url_for


# GET
@add_plone_route("/documents", "documents", methods=["GET"])
@add_plone_route("/documents/<string:uid>", "documents", methods=["GET"])
@returns_plone_items_for("documents")
def get(context, request, uid=None):
    """ get documents
    """
    return get_batched("Document", uid=uid, endpoint="documents")


# CREATE
@add_plone_route("/documents/create", "documents_create", methods=["POST"])
@add_plone_route("/documents/create/<string:uid>", "documents_create", methods=["POST"])
def create(context, request, uid=None):
    """ create documents
    """
    items = create_items("Document", uid=uid, endpoint="documents")
    return {
        "url": url_for("documents_create"),
        "count": len(items),
        "items": items,
    }


# UPDATE
@add_plone_route("/documents/update", "documents_update", methods=["POST"])
@add_plone_route("/documents/update/<string:uid>", "documents_update", methods=["POST"])
def update(context, request, uid=None):
    """ update documents
    """
    items = update_items("Document", uid=uid, endpoint="documents")
    return {
        "url": url_for("documents_update"),
        "count": len(items),
        "items": items,
    }


# DELETE
@add_plone_route("/documents/delete", "documents_delete", methods=["POST"])
@add_plone_route("/documents/delete/<string:uid>", "documents_delete", methods=["POST"])
def delete(context, request, uid=None):
    """ delete documents
    """
    items = delete_items("Document", uid=uid, endpoint="documents")
    return {
        "url": url_for("documents_delete"),
        "count": len(items),
        "items": items,
    }

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

__author__    = 'Ramon Bartl <ramon.bartl@googlemail.com>'
__docformat__ = 'plaintext'

import logging

from plone import api as ploneapi
from plone.jsonapi.core import router

from AccessControl import Unauthorized

from Products.CMFCore.interfaces import ISiteRoot
from Products.CMFCore.interfaces import IFolderish
from Products.ZCatalog.interfaces import ICatalogBrain
from Products.CMFPlone.PloneBatch import Batch

# search helpers
from query import search
from query import make_query

# request helpers
from plone.jsonapi.routes import request as req
from plone.jsonapi.routes.exceptions import APIError

from plone.jsonapi.routes.interfaces import IInfo
from plone.jsonapi.routes.interfaces import IBatch
from plone.jsonapi.routes.interfaces import IDataManager
from plone.jsonapi.routes import underscore as _


logger = logging.getLogger("plone.jsonapi.routes")

PORTAL_IDS = ["0", "portal", "site", "plone"]


#-----------------------------------------------------------------------------
#   Json API (CRUD) Functions
#-----------------------------------------------------------------------------


### GET RECORD
def get_record(uid=None):
    """ returns a single record
    """
    obj = None
    if uid is not None:
        obj = get_object_by_uid(uid)
    else:
        form = req.get_form()
        obj = get_object_by_record(form)
    if obj is None: raise APIError(404, "No object found")
    items = make_items_for([obj])
    return _.first(items)


### GET
def get_items(portal_type=None, request=None, uid=None, endpoint=None):
    """ returns a list of items

    1. If the UID is given, fetch the object directly => should return 1 item
    2. If no UID is given, search for all items of the given portal_type
    """

    # fetch the catalog results
    results = get_search_results(portal_type=portal_type, uid=uid)

    complete = req.get_complete()
    if not complete:
        # if the uid is given, get the complete information set
        complete = uid and True or False

    return make_items_for(results, endpoint, complete=complete)


### GET BATCHED
def get_batched(portal_type=None, request=None, uid=None, endpoint=None):
    """ returns a batched result record (dictionary)
    """

    # fetch the catalog results
    results = get_search_results(portal_type=portal_type, uid=uid)

    # fetch the batch params from the request
    size  = req.get_batch_size()
    start = req.get_batch_start()

    complete = req.get_complete()
    if not complete:
        # if the uid is given, get the complete information set
        complete = uid and True or False

    # return a batched record
    return get_batch(results, size, start, endpoint=endpoint, complete=complete)


### CREATE
def create_items(portal_type=None, request=None, uid=None, endpoint=None):
    """ create items

    1. If the uid is given, get the object and create the content in there
       (assumed that it is folderish)
    2. If the uid is 0, the target folder is assumed the portal.
    3. If there is no uid given, the payload is checked for either a key
        - `parent_uid`  specifies the *uid* of the target folder
        - `parent_path` specifies the *physical path* of the target folder
    """

    # destination where to create the content
    dest = uid and get_object_by_uid(uid) or None

    # extract the data from the request
    records = req.get_request_data()

    results = []
    for record in records:
        if dest is None:
            # find the container for content creation
            dest = find_target_container(record)
        if portal_type is None:
            portal_type = record.get("portal_type", None)
        id = record.get("id", None)
        title = record.get("title", None)
        obj = create_object_in_container(dest, portal_type, id=id, title=title)
        # update the object
        update_object_with_data(obj, record)
        results.append(obj)

    if not results:
        raise APIError(400, "No Objects could be created")

    return make_items_for(results, endpoint=endpoint)


### UPDATE
def update_items(portal_type=None, request=None, uid=None, endpoint=None):
    """ update items

    1. If the uid is given, the user wants to update the object with the data
       given in request body
    2. If no uid is given, the user wants to update a bunch of objects.
       -> each record contains either an UID, path or parent_path + id
    """

    # the data to update
    records = req.get_request_data()

    # we have an uid -> try to get an object for it
    obj = get_object_by_uid(uid)
    if obj:
        record = records[0] # ignore other records if we got an uid
        obj = update_object_with_data(obj, record)
        return make_items_for([obj], endpoint=endpoint)

    # no uid -> go through the record items
    results = []
    for record in records:
        obj = get_object_by_record(record)

        # no object found for this record
        if obj is None:
            continue

        # update the object with the given record data
        obj = update_object_with_data(obj, record)
        results.append(obj)

    if not results:
        raise APIError(400, "No Objects could be updated")

    return make_items_for(results, endpoint=endpoint)


### DELETE
def delete_items(portal_type=None, request=None, uid=None, endpoint=None):
    """ delete items

    1. If the uid is given, we can ignore the request body and delete the
       object with the given uid (if the uid was valid).
    2. If no uid is given, the user wants to delete more than one item.
       => go through each item and extract the uid. Delete it afterwards.
       // we should do this kind of transaction base. So if we can not get an
       // object for an uid, no item will be deleted.
    3. we could check if the portal_type matches, just to be sure the user
       wants to delete the right content.
    """

    # the data to update
    records = req.get_request_data()

    # we have an uid -> try to get an object for it
    obj = get_object_by_uid(uid)
    if obj:
        info = IInfo(obj)()
        info["deleted"] = delete_object(obj)
        return [info]

    # no uid -> go through the record items
    results = []
    for record in records:
        obj = get_object_by_record(record)

        # no object found for this record
        if obj is None:
            continue

        info = IInfo(obj)()
        info["deleted"] = delete_object(obj)
        results.append(info)

    if not results:
        raise APIError(400, "No Objects could be deleted")

    return results

#-----------------------------------------------------------------------------
#   Data Functions
#-----------------------------------------------------------------------------

def get_search_results(**kw):
    """ search the catalog and return the results

    The request may contain additional query parameters
    """
    query = make_query(**kw)
    return search(query)


def make_items_for(brains_or_objects, endpoint=None, complete=True):
    """ return a list of info dicts

    @param brains_or_objects: List/LazyMap
    """

    # this function extracts the data for one brain or object
    def _block(brain_or_object):

        if req.get("only_children"):
            return get_children(brain_or_object)

        # extract the data using the default info adapter
        info = IInfo(brain_or_object)()

        # might be None for mixed type catalog results, e.g. in the search route
        scoped_endpoint = endpoint
        if scoped_endpoint is None:
            scoped_endpoint = get_endpoint(get_portal_type(brain_or_object))

        info.update(get_url_info(brain_or_object, scoped_endpoint))

        # switch to wake up the object and complete the informations with the
        # data of the content adapter
        if complete:
            obj = get_object(brain_or_object)
            info.update(IInfo(obj)())
            info.update(get_parent_info(obj))
            if req.get_children():
                info.update(get_children(obj))

        return info

    return map(_block, brains_or_objects)


def get_url_info(brain_or_object, endpoint):
    """ returns the url info for the object
    """

    uid = get_uid(brain_or_object)
    return {
        "uid": uid,
        "url": get_url(brain_or_object),
        "api_url": url_for(endpoint, uid=uid),
    }

def get_parent_info(obj):
    """ returns the infos for the parent object
    """

    # special case for the portal route
    if is_root(obj): return {}

    parent   = get_parent(obj)
    endpoint = get_endpoint(parent.portal_type)

    if is_root(parent):
        return {
            "parent_id":  parent.getId(),
            "parent_uid": 0,
            "parent_url": url_for("portal"),
        }

    return {
        "parent_id":  parent.getId(),
        "parent_uid": get_uid(parent),
        "parent_url": url_for(endpoint, uid=get_uid(parent))
    }


def get_children(obj):
    """ returns the contents for this object
    """

    # ensure we have an object
    obj = get_object(obj)

    if is_folderish(obj) is False:
        return {
            "children": None,
        }

    children = []
    for content in obj.listFolderContents():
        endpoint = get_endpoint(get_portal_type(content))
        child = {
            "uid":     get_uid(content),
            "url":     get_url(content),
            "api_url": url_for(endpoint, uid=get_uid(content)),
        }
        child.update(IInfo(content)())
        children.append(child)

    return {
        "children": children
    }

#-----------------------------------------------------------------------------
#   Batching Helpers
#-----------------------------------------------------------------------------

def get_batch(sequence, size, start=0, endpoint=None, complete=False):
    """ create a batched result record out of a sequence (catalog brains)
    """

    # we call an adapter here to allow backwards compatibility hooks
    batch = IBatch(Batch(sequence, size, start))

    return {
        "pagesize": batch.get_pagesize(),
        "next":     batch.make_next_url(),
        "previous": batch.make_prev_url(),
        "page":     batch.get_pagenumber(),
        "pages":    batch.get_numpages(),
        "count":    batch.get_sequence_length(),
        "items":    make_items_for([b for b in batch.get_batch()], endpoint, complete=complete),
    }

#-----------------------------------------------------------------------------
#   Functional Helpers
#-----------------------------------------------------------------------------

def get_portal():
    """ get the Plone site
    """
    return ploneapi.portal.getSite()


def get_tool(name):
    """ get a tool by name
    """
    return ploneapi.portal.get_tool(name)


def get_portal_catalog():
    """ get the portal_catalog tool
    """
    return get_tool("portal_catalog")


def get_portal_reference_catalog():
    """ return reference_catalog tool
    """
    return get_tool("reference_catalog")


def get_portal_workflow():
    """ return portal_workflow tool
    """
    return get_tool("portal_workflow")


def is_brain(obj):
    """ checks if the object is a catalog brain
    """
    return ICatalogBrain.providedBy(obj)


def is_root(obj):
    """ checks if the object is the site root
    """
    return ISiteRoot.providedBy(obj)


def is_folderish(obj):
    """ checks if the object is folderish
    """
    return IFolderish.providedBy(obj)


def get_locally_allowed_types(obj):
    """ get the locally allowed types of this object
    """
    if not is_folderish(obj): return []
    method = getattr(obj, "getLocallyAllowedTypes", None)
    if not callable(method): return []
    return method()


def url_for(endpoint, **values):
    """ returns the api url
    """
    try:
        return router.url_for(endpoint, force_external=True, values=values)
    except:
        # XXX plone.jsonapi.core should catch the BuildError of Werkzeug and
        #     throw another error which can be handled here.
        logger.warn("Could not build API URL for endpoint '%s'. No route provider registered?" % endpoint)
        return None


def get_url(obj):
    """ get the absolute url for this object
    """
    if is_brain(obj):
        return obj.getURL()
    return obj.absolute_url()


def get_uid(obj):
    """ get the UID of the brain/object
    """
    if is_brain(obj):
        return obj.UID
    if is_root(obj):
        return 0
    return obj.UID()


def get_portal_type(obj):
    """ return the portal type of this object
    """
    return obj.portal_type


def get_object(brain_or_object):
    """ return the referenced object
    """
    if not is_brain(brain_or_object):
        return brain_or_object
    return brain_or_object.getObject()


def get_parent(brain_or_object):
    """ return the referenced object """
    obj = get_object(brain_or_object)
    return obj.aq_parent


def get_path(brain_or_object):
    """ return the physical path
    """
    if is_brain(brain_or_object):
        return brain_or_object.getPath()
    return "/".join(brain_or_object.getPhysicalPath())


def get_endpoint(portal_type):
    """ get the endpoint for this type """
    # handle portal types with dots
    portal_type = portal_type.split(".").pop()
    # remove whitespaces
    portal_type = portal_type.replace(" ", "")
    # lower and pluralize
    portal_type = portal_type.lower() + "s"

    return portal_type


def get_object_by_record(record):
    """ locate the object by record
    """

    # nothing to do here
    if not record: return None

    if record.get("uid"):
        return get_object_by_uid(record["uid"])
    if record.get("path"):
        return get_object_by_path(record["path"])
    if record.get("parent_path") and record.get("id"):
        path = "/".join([record["parent_path"], record["id"]])
        return get_object_by_path(path)

    logger.warn("get_object_by_record::No object found! record='%r'" % record)
    return None


def get_object_by_uid(uid):
    """ Fetches an object by uid
    """

    # nothing to do here
    if uid is None: return None

    # define uid 0 as the portal object
    if str(uid).lower() in PORTAL_IDS:
        return get_portal()

    # we try to find the object with both catalogs
    pc = get_portal_catalog()
    rc = get_portal_reference_catalog()

    # try to find the object with the reference catalog first
    obj = rc.lookupObject(uid)
    if obj:
        return obj

    # try to find the object with the portal catalog
    res = pc(dict(UID=uid))
    if len(res) > 1:
        raise APIError(400, "More than one object found for UID %s" % uid)
    if not res:
        return None

    return get_object(res[0])


def get_object_by_path(path):
    """ fetch the object by physical path
    """

    # nothing to do here
    if not path: return None

    pc = get_portal_catalog()
    portal = get_portal()
    portal_path = get_path(portal)

    if not path.startswith(portal_path):
        raise APIError(404, "Not a physical path inside the portal")

    if path == portal_path:
        return portal

    res = pc(path=dict(query=path, depth=0))
    if not res:
        return None
    return get_object(res[0])


def mkdir(path):
    """ creates a folder structure by a given path
    """

    container = get_portal()
    segments  = path.split("/")
    curpath   = None

    for n, segment in enumerate(segments):
        # skip the first element
        if not segment:
            continue

        curpath = "/".join(segments[:n+1])
        obj = get_object_by_path(curpath)

        if obj:
            container = obj
            continue

        if not is_folderish(container):
            raise APIError(400, "Object at %s is not a folder" % curpath)

        # create the folder on the go
        container = ploneapi.content.create(
                container, type="Folder", title=segment, save_id=True)

    return container


def find_target_container(record):
    """ find the target container for this record
    """

    parent_uid  = record.get("parent_uid")
    parent_path = record.get("parent_path")

    target = None

    # Try to find the target object
    if parent_uid:
        target = get_object_by_uid(parent_uid)
    elif parent_path:
        target = get_object_by_path(parent_path)
        # Issue 18
        if target is None:
            target = mkdir(parent_path)
    else:
        raise APIError(404, "No target UID/PATH information found")

    if not target:
        raise APIError(404, "No target container found")

    return target


def do_action_for(obj, transition):
    """ perform wf transition """
    return ploneapi.content.transition(obj, transition)


def delete_object(obj):
    """ delete the object """
    # we do not want to delete the site root!
    if is_root(obj):
        raise APIError(401, "Removing the Portal is not allowed")
    try:
        success = ploneapi.content.delete(obj) == None and True or False
    except Unauthorized:
        raise APIError(401, "You are not allowed to delete the object '%s'" % obj.getId())


def get_current_user():
    """ return the current logged in user """
    return ploneapi.user.get_current()


def create_object_in_container(container, portal_type, id=None, title=None):
    """ creates an object with the given data in the container
    """
    if not is_root(container) and portal_type not in get_locally_allowed_types(container):
        raise APIError(500, "Creation of this portal type is not allowed in this context.")
    return create_object(container=container, id=id, title=title, type=portal_type)


def create_object(**kw):
    defaults = {"save_id": True}
    defaults.update(kw)
    try:
        return ploneapi.content.create(**defaults)
    except Unauthorized:
        raise APIError(401, "You are not allowed to create this content")


def update_object_with_data(content, record):
    """ update the content with the values from records
    """
    dm = IDataManager(content, None)
    if dm is None:
        raise APIError(400, "Update on this object is not allowed")
    for k, v in record.items():
        try:
            success = dm.set(k, v)
        except Unauthorized:
            raise APIError(401, "You are not allowed to set the field '%s'" % k)

        if not success:
            logger.info("update_object_with_data::skipping key=%r", k)
            continue

        logger.info("update_object_with_data::field %r updated with value=%r", k, v)

    # do a wf transition
    if record.get("transition", None):
        t = record.get("transition")
        logger.info(">>> Do Transition '%s' for Object %s", t, content.getId())
        do_action_for(content, t)

    # reindex the object
    content.reindexObject()
    return content

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

from zope import interface

__author__ = 'Ramon Bartl <ramon.bartl@googlemail.de>'
__docformat__ = 'plaintext'


class IInfo(interface.Interface):
    """ JSON Info Interface
    """

    def to_dict():
        """ return the dictionary representation of the object
        """

    def __call__():
        """ return the dictionary representation of the object
        """


class IBatch(interface.Interface):
    """ Batch Interface
    """

    def get_batch():
        """ return the wrapped batch object
        """

    def get_pagesize():
        """ return the current page size
        """

    def get_pagenumber():
        """ return the current page number
        """

    def get_numpages():
        """ return the current number of pages
        """

    def get_sequence_length():
        """ return the length
        """

    def make_next_url():
        """ build and return the next url
        """

    def make_prev_url():
        """ build and return the previous url
        """


class IDataManager(interface.Interface):
    """ Field Interface
    """

    def get(name):
        """ Get the value of the named field with
        """

    def set(name, value):
        """ Set the value of the named field
        """



# -*- coding: utf-8 -*-

import types
import logging
import pkg_resources

from plone import api
from ZPublisher import HTTPRequest

from DateTime import DateTime

from plone.jsonapi.routes import request as req
from plone.jsonapi.routes import underscore as _

try:
    pkg_resources.get_distribution('Products.AdvancedQuery')
    from Products.AdvancedQuery import Eq
    from Products.AdvancedQuery import In
    from Products.AdvancedQuery import Generic
except pkg_resources.DistributionNotFound:
    HAS_ADVANCED_QUERY = False
else:
    HAS_ADVANCED_QUERY = True

__author__ = 'Ramon Bartl <ramon.bartl@googlemail.com>'
__docformat__ = 'plaintext'


__all__ = ['search', 'make_query']


logger = logging.getLogger("plone.jsonapi.routes.query")

# TODO: make this an configurable option -- maybe zcml?
USE_ADVANCED_QUERY = False


# -----------------------------------------------------------------------------
#   Public API
# -----------------------------------------------------------------------------

def search(query, **kw):
    """ execute either AdvancedQuery or a StandardQuery
    """
    if USE_ADVANCED_QUERY and HAS_ADVANCED_QUERY and type(query) is tuple:
        return advanced_search(*query, **kw)
    return standard_search(query, **kw)


def make_query(**kw):
    """ generates a catalog query
    """
    if USE_ADVANCED_QUERY and HAS_ADVANCED_QUERY:
        return make_advanced_query(**kw)
    return make_standard_query(**kw)


# -----------------------------------------------------------------------------
#   Query Builders
# -----------------------------------------------------------------------------

def make_standard_query(**kw):
    """ generates a query for the portal catalog
    """
    logger.info("Building **standard** query")

    # build a default query from the request parameters and the keywords
    query = build_catalog_query(**kw)

    sort_on, sort_order = get_sort_spec()
    query.update(dict(sort_order=sort_order, sort_on=sort_on))

    return query


def make_advanced_query(**kw):
    """ generates a query suitable for the portal catalog
    """
    logger.info("Building **advanced** query")

    # build the initial query
    query = build_catalog_query(**kw)

    # transform to advanced query
    advanced_query = to_advanced_query(query)

    # get the sort specification
    sort = get_sort_spec()

    return advanced_query, (sort, )


def build_catalog_query(**kw):
    """ build an initial query object

    this query can be used directly for a std. catalog query
    """
    query = {}

    # note: order is important!
    query.update(get_request_query())
    query.update(get_custom_query())
    query.update(get_keyword_query(**kw))

    logger.info("build_catalog_query::query=%s" % query)
    return query


def get_request_query():
    """ checks the request for known catalog indexes.
    """
    query = {}

    # only known indexes get observed
    indexes = get_catalog_indexes()

    # check what we can use from the reqeust
    request = req.get_request()

    for idx in indexes:
        val = request.form.get(idx)
        if val:
            query[idx] = to_index_value(val, idx)

    return query


def get_custom_query():
    """ checks the request for custom query keys.
    """
    query = {}

    # searchable text queries
    q = req.get_query()
    if q:
        query["SearchableText"] = q

    # physical path queries
    path = req.get_path()
    if path:
        query["path"] = {'query': path, 'depth': req.get_depth()}

    # special handling for recent created/modified
    recent_created = req.get_recent_created()
    if recent_created:
        date = calculate_delta_date(recent_created)
        query["created"] = {'query': date, 'range': 'min'}

    recent_modified = req.get_recent_modified()
    if recent_modified:
        date = calculate_delta_date(recent_modified)
        query["modified"] = {'query': date, 'range': 'min'}

    return query


def get_keyword_query(**kw):
    """ generates a query from the given keywords
    """
    query = dict()

    # only known indexes get observed
    indexes = get_catalog_indexes()

    for k, v in kw.iteritems():
        # handle uid
        if v and k.lower() == "uid":
            if v:
                query["UID"] = v
            continue
        # handle portal_type
        if k.lower() == "portal_type":
            if v:
                query["portal_type"] = _.to_list(v)
            continue
        # and the rest
        if k not in indexes:
            logger.warn("Skipping unknown keyword parameter '%s=%s'" % (k, v))
            continue
        if v is None:
            logger.warn("Skip value 'None' in kw parameter '%s=%s'" % (k, v))
            continue
        logger.info("Adding '%s=%s' to query" % (k, v))
        query[k] = v

    return query


def to_advanced_query(query):
    """ convert a dictionary to an advanced query
    """

    # nothing to do
    if not query:
        return Eq("Title", "")

    a_query = None

    def get_query_expression_for(value):
        # return the Advanced Query Expression
        if type(value) in (tuple, list):
            return In
        if type(value) is dict:
            return Generic
        return Eq

    for k, v in query.iteritems():
        exp = get_query_expression_for(v)
        # first loop, build the initial query expression
        if a_query is None:
            a_query = exp(k, v)
        else:
            a_query = a_query & exp(k, v)

    return a_query


# -----------------------------------------------------------------------------
#   Functional Helpers
# -----------------------------------------------------------------------------

def calculate_delta_date(literal):
    """ calculate the date in the past from the given literal
    """
    mapping = {
        "today":      0,
        "yesterday":  1,
        "this-week":  7,
        "this-month": 30,
        "this-year":  365,
    }
    today = DateTime(DateTime().Date())  # current date without the time
    return today - mapping.get(literal, 0)


# -----------------------------------------------------------------------------
#   Catalog Helpers
# -----------------------------------------------------------------------------

def get_portal_catalog():
    """ fetch portal_catalog tool
    """
    return api.portal.get_tool("portal_catalog")


def get_catalog_indexes():
    """ return the list of indexes of the portal catalog
    """
    return get_portal_catalog().indexes()


def get_index(name):
    """ get the index object by name
    """
    return get_portal_catalog()._catalog.getIndex(name)


def to_index_value(value, index):
    """ convert the value for the given index
    """
    # ZPublisher records can be passed to the catalog as is.
    if isinstance(value, HTTPRequest.record):
        return value

    if type(index) in types.StringTypes:
        index = get_index(index)

    if index.meta_type == "DateIndex":
        return DateTime(value)
    if index.meta_type == "BooleanIndex":
        return bool(value)
    if index.meta_type == "KeywordIndex":
        return value.split(",")

    return value


def get_sort_spec():
    """ build sort specification
    """
    all_indexes = get_portal_catalog().indexes()
    si = req.get_sort_on(allowed_indexes=all_indexes)
    so = req.get_sort_order()
    return si, so


def standard_search(query, **kw):
    """ search the portal catalog
    """
    logger.info("Standard Query -> %r" % (query))
    pc = get_portal_catalog()
    return pc(query, **kw)


def advanced_search(query, sort_specs=()):
    """ search the portal catalog with the advanced query
    """
    logger.info("Advanced Query -> %s sort_specs=%r" % (query, sort_specs))
    pc = get_portal_catalog()
    return pc.evalAdvancedQuery(query, sort_specs)



# -*- coding: utf-8 -*-

import os
import re
import datetime
import fileinput

# Fabric imports
from fabric.api import env
from fabric.api import lcd
from fabric.api import task
from fabric.api import local
from fabric.api import execute
from fabric.colors import red
from fabric.colors import green
from fabric.colors import yellow

if not os.path.exists("fabfile.py"):
    raise RuntimeError("Must be run in the buildout directory")

env.version_file = "src/plone/jsonapi/routes/version.py"


# -----------------------------------------------------------------------------
# Public API
# -----------------------------------------------------------------------------

@task
def test():
    local("./bin/test -s plone.jsonapi.routes")


@task
def make_docs():
    with lcd("docs"):
        local("make html")


@task
def preview_docs():
    with lcd("docs"):
        local("open _build/html/index.html")


@task
def bump_version():
    """ Bump up the version number
    """
    print yellow("bumping version ...")
    write_version_info()
    print red("adding file '%s' to next commit." % env.version_file)
    local("git add " + env.version_file)
    print "updated version: ", version_string()


@task
def version():
    """ Print current version
    """
    print green(version_string())


@task
def reload():
    execute("bump_version")
    local("wget --delete-after http://admin:admin@127.0.0.1:8080/@@reload?action=code")


# -----------------------------------------------------------------------------
# Functional Helpers
# -----------------------------------------------------------------------------

def write_version_info():
    """ updates the build and date of the version module
    """
    f = env.version_file
    if not os.path.exists(f):
        raise RuntimeError("File '%s' does not exist." % f)

    for line in fileinput.input(f, inplace=1):
        if re.findall("__build__.*=", line):
            build = int(line.split("=")[1])
            line = "__build__ = %d" % (build + 1)
        elif re.findall("__date__.*=", line):
            now = datetime.datetime.now().strftime("%Y-%m-%d")
            line = "__date__ = '%s'" % now
        print line.strip("\n")


def get_version():
    f = env.version_file
    out = {}
    lines = file(f).readlines()
    for l in lines:
        if "=" in l and l.split("=")[0].strip() in ("__build__", "__date__", "__version__"):
            name, value = l.split("=")
            name = name.strip(" _\n\r")
            value = value.strip(' "\'\n\r')
            out[name] = value
    return out


def version_string(version=None):
    if not version:
        version = get_version()
    return "version %(version)s build %(build)s date %(date)s" % (version)



# -*- coding: utf-8 -*-

from plone.jsonapi.routes import add_plone_route

# CRUD
from plone.jsonapi.routes.api import get_batched
from plone.jsonapi.routes.api import create_items
from plone.jsonapi.routes.api import update_items
from plone.jsonapi.routes.api import delete_items

from plone.jsonapi.routes.api import url_for


# GET
@add_plone_route("/newsitems", "newsitems", methods=["GET"])
@add_plone_route("/newsitems/<string:uid>", "newsitems", methods=["GET"])
def get(context, request, uid=None):
    """ get newsitems
    """
    return get_batched("News Item", uid=uid, endpoint="newsitems")


# CREATE
@add_plone_route("/newsitems/create", "newsitems_create", methods=["POST"])
@add_plone_route("/newsitems/create/<string:uid>", "newsitems_create", methods=["POST"])
def create(context, request, uid=None):
    """ create newsitems
    """
    items = create_items("News Item", uid=uid, endpoint="newsitems")
    return {
        "url": url_for("newsitems_create"),
        "count": len(items),
        "items": items,
    }


# UPDATE
@add_plone_route("/newsitems/update", "newsitems_update", methods=["POST"])
@add_plone_route("/newsitems/update/<string:uid>", "newsitems_update", methods=["POST"])
def update(context, request, uid=None):
    """ update newsitems
    """
    items = update_items("News Item", uid=uid, endpoint="newsitems")
    return {
        "url": url_for("newsitems_update"),
        "count": len(items),
        "items": items,
    }


# DELETE
@add_plone_route("/newsitems/delete", "newsitems_delete", methods=["POST"])
@add_plone_route("/newsitems/delete/<string:uid>", "newsitems_delete", methods=["POST"])
def delete(context, request, uid=None):
    """ delete newsitems
    """
    items = delete_items("News Item", uid=uid, endpoint="newsitems")
    return {
        "url": url_for("newsitems_delete"),
        "count": len(items),
        "items": items,
    }

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

__author__    = 'Ramon Bartl <ramon.bartl@googlemail.com>'
__docformat__ = 'plaintext'


from plone.jsonapi.routes import request as req


class APIError(Exception):
    """ Exception Class for API Errors
    """

    def __init__(self, status, message):
        self.message = message
        self.status = status
        self.setStatus(status)

    def setStatus(self, status):
        request = req.getRequest()
        request.response.setStatus(status)

    def __str__(self):
        return self.message

# vim: set ft=python ts=4 sw=4 expandtab :



##########################################################
#
# Licensed under the terms of the GNU Public License
# (see docs/LICENSE.GPL)
#
# Copyright (c) 2005:
#   - The Open Planning Project (http://www.openplans.org/)
#   - Whit Morriss <whit@kalistra.com>
#   - Rob Miller <rob@kalistra.com> (RaFromBRC)
#   - and contributors
#
##########################################################
from interfaces import IAmWickedField, IAmWicked, IFieldEvent
from interfaces import ICacheManager, IValueToString, IScope
from interfaces import IWickedFilter, IWickedQuery, IBacklinkManager
from normalize import titleToNormalizedId as normalize
from wicked import utils
from wicked.fieldevent.interfaces import EndFiltrationException
from wicked.fieldevent.interfaces import ITxtFilterList, IFieldRenderEvent
from wicked.fieldevent.interfaces import IFieldValueSetter, IFieldStorageEvent
from wicked.fieldevent.interfaces import IFieldEvent
from zope.component.interfaces import ComponentLookupError
from wicked.fieldevent.txtfilter import TxtFilter
from zope.component import getMultiAdapter, adapts, adapter
from zope.interface import implements, implementer, Interface, alsoProvides

import re

_marker = object()

pattern1 = re.compile(r'\(\(([\w\W]+?)\)\)') # matches ((Some Text To link 123))
pattern2 = re.compile(r'\[\[([\w\W]+?)\]\]') # matches [[Some Text To link 123]]

def removeParens(wikilink):
    wikilink.replace('((', '')
    wikilink.replace('))', '')
    wikilink.replace('[[', '')
    wikilink.replace(']]', '')
    return wikilink

class WickedFilter(TxtFilter):
    implements(IWickedFilter)
    adapts(IAmWickedField, IAmWicked, IFieldEvent)

    name = 'Wicked Filter'

    #pattern = [pattern1, pattern2]
    query_iface = IWickedQuery
    _encoding = 'UTF8'

    def __init__(self, field, instance, event):
        super(WickedFilter, self).__init__(field, instance, event)
        self.section = field.__name__
        self.pattern = None

    @utils.memoizedproperty
    def scope(self):
        try:
            return getMultiAdapter((self.field, self.context), IScope)
        except ComponentLookupError:
            return ''

    # avoid global lookup
    getMatch = staticmethod(utils.getMatch)
    _normalize = staticmethod(normalize)

    # optimization
    @utils.memoize
    def normalize(self, value):
        return self._normalize(value)

    @utils.memoizedproperty
    def encoding(self):
        """AT hack"""
        try:
            encoding = self.context.getCharset()
        except AttributeError:
            encoding = self._encoding
        return encoding

    def _filterCore(self, chunk, **kwargs):
        normalled = self.normalize(chunk)
        links=self.getLinks(chunk, normalled)
        self.renderer.load(links, chunk)
        return self.renderer().encode(self.encoding)

    @property
    def filtered_text(self):
        """syntax preprocessing"""
        return super(WickedFilter, self).filtered_text

    @utils. memoize
    @utils.linkcache
    def getLinks(self, chunk, normalled):
        self.resolver.configure(chunk, normalled, self.scope)
        brains = self.resolver.search
        if not brains:
            brains = self.resolver.scopedSearch
        links = [utils.packBrain(b) for b in brains if b]
        return links

    @utils.memoizedproperty
    def resolver(self):
        """
        @return query object
        """
        return self.query_iface(self.context)

    @utils.memoizedproperty
    def backlinker(self):
        return getMultiAdapter((self, self.context), IBacklinkManager)

    def manageLink(self, obj, link):
        self.backlinker.manageLink(obj, link)

    def unlink(self, uid):
        self.backlinker.unlink(uid)

    def manageLinks(self, links):
        self.backlinker.manageLinks(links)

    @utils.memoizedproperty
    def cache(self):
        return getMultiAdapter((self, self.context), ICacheManager)

    @utils.memoizedproperty
    def renderer(self):
        # @@ better way to get request? maybe a txtfilter should be a view?
        renderer = getMultiAdapter((self.context, self.context.REQUEST), Interface, 'link_renderer')
        renderer.section = self.section
        # hook for zope2 aq wrapper
        if hasattr(renderer, '__of__'):
            return renderer.__of__(self.context)
        return renderer

    def __call__(self):
        if self.event.kwargs.get('raw', False):
            raise EndFiltrationException('Kwargs flag for raw return')
        super(WickedFilter, self).__call__()

##     def removeParens(wikilink):
##         wikilink.replace('((', '')
##         wikilink.replace('))', '')
##         return wikilink
    removeParens=staticmethod(removeParens)


class BrackettedWickedFilter(WickedFilter):
    """media wiki style bracket matching"""
    pattern=pattern2
    def removeParens(wikilink):
        wikilink.replace('[[', '')
        wikilink.replace(']]', '')
        return wikilink
    removeParens=staticmethod(removeParens)

NAME = WickedFilter.name


## event handlers ##

class WickedListener(object):

    def __init__(self, pattern):
        self.pattern = pattern

    def render(self, field, instance, event):
        """standalone wicked filter (ie not as a txtfilter). Optimal if
        not using txtfilters"""

        if event.kwargs.get('raw', False):
            return

        wicked = getMultiAdapter((field, instance, event), IWickedFilter)
        wicked.pattern = self.pattern
        try:
            wicked()
        except EndFiltrationException:
            pass

    def store(self, field, event):
        try:
            wicked = utils.getWicked(field, event.instance, event)
        except ComponentLookupError:
            # no adapter registered for this type currently
            # @@ This might be handle better by redispatch
            return

        wicked.pattern = self.pattern
        if not event.value:
            return

        value = event.value
        value_str = value

        try:
            # this block handle conversions for file uploads and
            # atapi.BaseUnit or any other not quite plain text "value objects"
            value_str = getMultiAdapter((value, field), IValueToString)
        except ComponentLookupError:
            pass

        found = wicked.findall(value_str)

        if not len(found):
            return

        new_links = [wicked.removeParens(link) for link in found]
        wicked.manageLinks(new_links)

from zope.interface import classImplements

@implementer(IFieldValueSetter)
@adapter(IAmWickedField, IFieldStorageEvent)
def backlink_handler(field, event):
    try:
        wicked = utils.getWicked(field, event.instance, event)
    except ComponentLookupError:
        # no adapter registered for this type currently
        # @@ This might be handle better by redispatch
        return

    if not event.value:
        return

    value = event.value
    value_str = value

    try:
        # this block handle conversions for file uploads and
        # atapi.BaseUnit or any other not quite plain text "value objects"
        value_str = getMultiAdapter((value, field), IValueToString)
    except ComponentLookupError:
        pass


    found = wicked.findall(value_str)

    if not len(found):
        return

    new_links = [wicked.removeParens(link) for link in found]
    wicked.manageLinks(new_links)

pattern1_listeners = WickedListener(pattern1)
pattern2_listeners = WickedListener(pattern2)

## hack around fact that functions can not be pickled ##

class wicked_listener(object):
    __init__=staticmethod(pattern1_listeners.render)

class bracketted_wicked_listener(object):
    __init__=staticmethod(pattern2_listeners.render)

class backlink(object):
    implements(IFieldValueSetter)
    adapts(IAmWickedField, IFieldStorageEvent)

    __init__ = staticmethod(pattern1_listeners.store)

BacklinkRegistrationProxy = backlink

class brackettedbacklink(object):
    implements(IFieldValueSetter)
    adapts(IAmWickedField, IFieldStorageEvent)

    __init__ = staticmethod(pattern2_listeners.store)


## toy example code ##

@adapter(IAmWickedField, IAmWicked, IFieldRenderEvent)
@implementer(ITxtFilterList)
def filter_list(field, context, event):
    """example adapter for a one item list for ordering a txtfilter
    pipeline involving wicked only.  Practically useless, for example
    only"""
    return [NAME]



# -*- coding: utf-8 -*-

from plone.jsonapi.routes import add_plone_route

# CRUD
from plone.jsonapi.routes.api import get_batched


@add_plone_route("/search", "search", methods=["GET"])
def get(context, request):
    """ search all contents
    """
    return get_batched([])

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

import simplejson as json

import unittest2 as unittest

from plone.testing.z2 import Browser

from plone.app.testing import PloneSandboxLayer
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.testing import z2


from plone.app.testing.layers import IntegrationTesting
from zope.configuration import xmlconfig


class TestLayer(PloneSandboxLayer):
    defaultBases = (PLONE_FIXTURE,)


    def setUpZope(self, app, configurationContext):
        # Load ZCML
        import plone.jsonapi.core
        import plone.jsonapi.routes
        xmlconfig.file('configure.zcml', plone.jsonapi.core, context=configurationContext)
        xmlconfig.file('configure.zcml', plone.jsonapi.routes, context=configurationContext)

        # Install product and call its initialize() function
        z2.installProduct(app, 'plone.jsonapi.core')
        z2.installProduct(app, 'plone.jsonapi.routes')

    def tearDownZope(self, app):
        # Uninstall product
        z2.uninstallProduct(app, 'plone.jsonapi.core')
        z2.uninstallProduct(app, 'plone.jsonapi.routes')

    def setUpPloneSite(self, portal):
        setRoles(portal, TEST_USER_ID, ['Manager'])

        # add a folder, so we can test with it
        _ = portal.invokeFactory("Folder", "folder", title="Test Folder")
        folder = portal[_]
        for i in range(50):
            folder.invokeFactory("Document", "document-%d" % i, title="Test Document %d" % i)

        # Test fixture -- p.j.c. needs to have a request
        from plone.jsonapi.core import router
        router.DefaultRouter.initialize(portal, portal.REQUEST)


TEST_FIXTURE = TestLayer()
INTEGRATION_TESTING = IntegrationTesting(bases=(TEST_FIXTURE,),
                          name="plone.jsonapi.routes:Integration")


class APITestCase(unittest.TestCase):
    layer = INTEGRATION_TESTING

    def getBrowser(self, handleErrors=False):
        browser = Browser(self.getApp())
        if handleErrors:
            browser.handleErrors = True
        return browser

    def getApp(self):
        return self.layer.get("app")

    def getPortal(self):
        return self.layer.get("portal")

    def getRequest(self):
        return self.layer.get("request")

    def decode(self, s):
        return json.loads(s)

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

from plone.jsonapi.core import router

from plone.jsonapi.routes.tests.base import APITestCase


class TestSetup(APITestCase):
    """ Test URL registration machinery
    """

    def testCRUDEndpoint(self):
        endpoint = "get"
        self.assertEqual(router.url_for(endpoint), "/plone/@@API/plone/api/1.0/%s" % endpoint)

    def testCollectionsEndpoint(self):
        endpoint = "collections"
        self.assertEqual(router.url_for(endpoint), "/plone/@@API/plone/api/1.0/%s" % endpoint)

    def testDocumentsEndpoint(self):
        endpoint = "documents"
        self.assertEqual(router.url_for(endpoint), "/plone/@@API/plone/api/1.0/%s" % endpoint)

    def testEventsEndpoint(self):
        endpoint = "events"
        self.assertEqual(router.url_for(endpoint), "/plone/@@API/plone/api/1.0/%s" % endpoint)

    def testFilesEndpoint(self):
        endpoint = "files"
        self.assertEqual(router.url_for(endpoint), "/plone/@@API/plone/api/1.0/%s" % endpoint)

    def testFoldersEndpoint(self):
        endpoint = "folders"
        self.assertEqual(router.url_for(endpoint), "/plone/@@API/plone/api/1.0/%s" % endpoint)

    def testImagesEndpoint(self):
        endpoint = "images"
        self.assertEqual(router.url_for(endpoint), "/plone/@@API/plone/api/1.0/%s" % endpoint)

    def testLinksEndpoint(self):
        endpoint = "links"
        self.assertEqual(router.url_for(endpoint), "/plone/@@API/plone/api/1.0/%s" % endpoint)

    def testNewsItemsEndpoint(self):
        endpoint = "newsitems"
        self.assertEqual(router.url_for(endpoint), "/plone/@@API/plone/api/1.0/%s" % endpoint)

    def testSearchEndpoint(self):
        endpoint = "search"
        self.assertEqual(router.url_for(endpoint), "/plone/@@API/plone/api/1.0/%s" % endpoint)

    def testTopicEndpoint(self):
        endpoint = "topics"
        self.assertEqual(router.url_for(endpoint), "/plone/@@API/plone/api/1.0/%s" % endpoint)

    def testUsersEndpoint(self):
        endpoint = "users"
        self.assertEqual(router.url_for(endpoint), "/plone/@@API/plone/api/1.0/%s" % endpoint)


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestSetup))
    return suite

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

__author__    = 'Ramon Bartl <ramon.bartl@googlemail.com>'
__docformat__ = 'plaintext'

import urllib
import logging

from zope import interface

from plone.jsonapi.routes.interfaces import IBatch

from plone.jsonapi.routes import request as req

logger = logging.getLogger("plone.jsonapi.routes.batching")


class Batch(object):
    """ Adapter for Plone 4.3 batching functionality
    """
    interface.implements(IBatch)

    def __init__(self, batch):
        self.batch = batch

    def get_batch(self):
        return self.batch

    def get_pagesize(self):
        return self.batch.pagesize

    def get_pagenumber(self):
        return self.batch.pagenumber

    def get_numpages(self):
        return self.batch.numpages

    def get_sequence_length(self):
        return self.batch.sequence_length

    def make_next_url(self):
        if not self.batch.has_next:
            return None
        request = req.get_request()
        params = request.form
        params["b_start"] = self.batch.pagenumber * self.batch.pagesize
        return "%s?%s" % (request.URL, urllib.urlencode(params))

    def make_prev_url(self):
        if not self.batch.has_previous:
            return None
        request = req.get_request()
        params = request.form
        params["b_start"] = max(self.batch.pagenumber - 2, 0) * self.batch.pagesize
        return "%s?%s" % (request.URL, urllib.urlencode(params))


class Batch42(object):
    """ Adapter for Plone 4.2 batching functionality
    """
    interface.implements(IBatch)

    def __init__(self, batch):
        self.batch = batch

    def get_batch(self):
        return self.batch

    def get_pagesize(self):
        return self.batch.size

    def get_pagenumber(self):
        return self.batch.pagenumber

    def get_numpages(self):
        return self.batch.numpages

    def get_sequence_length(self):
        return self.batch.sequence_length

    def make_next_url(self):
        if self.batch.next is not None:
            return None
        request = req.get_request()
        params = request.form
        params["b_start"] = self.batch.numpages * self.batch.size
        return "%s?%s" % (request.URL, urllib.urlencode(params))

    def make_prev_url(self):
        if self.batch.previous is not None:
            return None
        request = req.get_request()
        params = request.form
        params["b_start"] = max(self.batch.numpages - 2, 0) * self.batch.size
        return "%s?%s" % (request.URL, urllib.urlencode(params))

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

import json

from plone.jsonapi.routes.tests.base import APITestCase
from plone.jsonapi.routes import request as req


class TestRequest(APITestCase):
    """ Test the request API
    """

    def setUp(self):
        self.portal = self.getPortal()
        self.request = self.getRequest()

    def test_complete_parameter(self):
        self.assertFalse(req.get_complete())
        request = req.get_request()
        request.form["complete"] = "yes"
        self.assertTrue(req.get_complete())

    def test_sort_limit_parameter(self):
        self.assertEqual(req.get_sort_limit(), None)
        request = req.get_request()
        request.form["sort_limit"] = 1
        self.assertEqual(req.get_sort_limit(), 1)

    def test_batch_size_parameter(self):
        self.assertEqual(req.get_batch_size(), 25)
        request = req.get_request()
        request.form["limit"] = 1
        self.assertEqual(req.get_batch_size(), 1)

    def test_batch_start_parameter(self):
        self.assertEqual(req.get_batch_start(), 0)
        request = req.get_request()
        request.form["b_start"] = 1
        self.assertEqual(req.get_batch_start(), 1)

    def test_sort_on_parameter(self):
        self.assertEqual(req.get_sort_on(), None)
        request = req.get_request()
        request.form["sort_on"] = "Title"
        self.assertEqual(req.get_sort_on(), "Title")

    def test_sort_order_parameter(self):
        self.assertEqual(req.get_sort_order(), "ascending")
        request = req.get_request()
        request.form["sort_order"] = "desc"
        self.assertEqual(req.get_sort_order(), "descending")
        request.form["sort_order"] = "DESC"
        self.assertEqual(req.get_sort_order(), "descending")

    def test_query_parameter(self):
        self.assertEqual(req.get_query(), "")
        request = req.get_request()
        request.form["q"] = "#*!$Foo"
        self.assertEqual(req.get_query(), "Foo*")

    def test_path_parameter(self):
        self.assertEqual(req.get_path(), "")
        request = req.get_request()
        request.form["path"] = "/Plone/folder"
        self.assertEqual(req.get_path(), "/Plone/folder")

    def test_depth_parameter(self):
        self.assertEqual(req.get_depth(), 0)
        request = req.get_request()
        request.form["depth"] = 5
        self.assertEqual(req.get_depth(), 5)

    def test_recent_modified_parameter(self):
        self.assertEqual(req.get_recent_modified(), None)
        request = req.get_request()
        request.form["recent_modified"] = "this-week"
        self.assertEqual(req.get_recent_modified(), "this-week")

    def test_recent_created_parameter(self):
        self.assertEqual(req.get_recent_created(), None)
        request = req.get_request()
        request.form["recent_created"] = "this-week"
        self.assertEqual(req.get_recent_created(), "this-week")

    def test_request_data(self):
        self.assertEqual(req.get_request_data(), [{}])
        data = {"title": "test"}
        self.request["BODY"] = json.dumps(data)
        self.assertEqual(req.get_request_data(), [data])


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestRequest))
    return suite

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages

version = '0.6'

long_description = (
    open('README.rst').read()
    + '\n' +
    open(os.path.join('docs', 'HISTORY.rst')).read()
    + '\n')

setup(name='plone.jsonapi.routes',
      version=version,
      description="Plone JSON API -- Routes",
      long_description=long_description,
      # Get more strings from
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
        "Programming Language :: Python",
        "Framework :: Plone",
        "Framework :: Zope2",
        ],
      keywords='',
      author='Ramon Bartl',
      author_email='ramon.bartl@googlemail.com',
      url='https://github.com/ramonski/plone.jsonapi',
      license='MIT',
      packages=find_packages('src'),
      package_dir = {'': 'src'},
      namespace_packages=['plone', 'plone.jsonapi'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'plone.api',
          'plone.jsonapi.core>=0.3',
          # -*- Extra requirements: -*-
      ],
      extras_require={
          'test': [
               'plone.app.testing',
               'unittest2',
           ]
      },
      entry_points="""
      # -*- Entry points: -*-
      [z3c.autoinclude.plugin]
      target = plone
      """,
      )

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

__author__    = 'Ramon Bartl <ramon.bartl@googlemail.com>'
__docformat__ = 'plaintext'

import re
import logging

from plone.jsonapi.core.browser.router import add_route
from plone.jsonapi.core.browser.router import DefaultRouter


BASE_URL = "/plone/api/1.0"

logger = logging.getLogger("plone.jsonapi.routes")


def add_plone_route(rule, endpoint=None, **kw):
    """ add a Plone JSON API route
    """
    def apiurl(rule):
         return '/'.join(s.strip('/') for s in ["", BASE_URL, rule])

    def wrapper(f):
        try:
            DefaultRouter.add_url_rule(apiurl(rule), endpoint=endpoint, view_func=f, options=kw)
        except AssertionError:
            pass
        return f

    return wrapper


def get_api_routes_for(segment):
    """ return a list of all routes registered for the given url segment
    """
    adapter = DefaultRouter.get_adapter()

    out = []
    rx = re.compile(r".*%s/[\w]+$" % segment)

    for rule in adapter.map.iter_rules():
        if rx.match(rule.rule):
            endpoint = rule.endpoint
            info = DefaultRouter.view_functions.get(endpoint).__doc__
            url  = DefaultRouter.url_for(endpoint, force_external=True)
            out.append({
                "url":  url,
                "info": info and info.strip() or "No description available",
                "methods": ",".join(rule.methods),
            })
    return out


@add_route(BASE_URL, "api", methods=["GET"])
@add_plone_route("api.json", "api", methods=["GET"])
def api_json(context, request):
    """ API URLs
    """
    items = get_api_routes_for(BASE_URL)

    return {
        "url": DefaultRouter.url_for("api", force_external=True),
        "count": len(items),
        "items": items,
    }


def initialize(context):
    """ Initializer called when used as a Zope 2 product.
    """
    logger.info("*** PLONE.JSONAPI.ROUTES INITIALIZE ***")

    # We have to import the modules so that the routes get initialized
    import pkgutil
    import providers

    prefix = providers.__name__ + "."
    for importer, modname, ispkg in pkgutil.iter_modules(providers.__path__, prefix):
        module = __import__(modname, fromlist="dummy")
        logger.info("INITIALIZED ROUTE PROVIDER ---> %s" % module.__name__)

    import version
    logger.info("INITIALIZED ROUTE PROVIDER ---> %s" % version.__name__)

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

from plone import api as ploneapi

from plone.jsonapi.routes import add_plone_route
from plone.jsonapi.routes.api import url_for


def get_user_info(username=None, short=True):
    """ return the user informations
    """

    # XXX: refactoring needed in this function

    user = None
    anon = ploneapi.user.is_anonymous()
    current = ploneapi.user.get_current()

    # no username, go and get the current user
    if username is None:
        user = current
    else:
        user = ploneapi.user.get(username)

    if not user:
        raise KeyError('User not found')

    info = {
        "id":       user.getId(),
        "username": user.getUserName(),
        "url":      url_for("users", username=user.getUserName())
    }

    # return base info
    if short or anon:
        return info

    # try to get extended infos
    pu = user.getUser()
    properties = {}
    if "mutable_properties" in pu.listPropertysheets():
        mp = pu.getPropertysheet("mutable_properties")
        properties = dict(mp.propertyItems())

    def to_iso8601(dt=None):
        if dt is None:
            return ""
        return dt.ISO8601()

    # include mutable properties if short==False
    info.update({
        "email":           properties.get("email"),
        "fullname":        properties.get("fullname"),
        "login_time":      to_iso8601(properties.get("login_time")),
        "last_login_time": to_iso8601(properties.get("last_login_time")),
        "roles":           user.getRoles(),
        "groups":          pu.getGroups(),
        "authenticated":   current == user and not anon,
    })

    return info


#-----------------------------------------------------------------------------
# API ROUTES
#-----------------------------------------------------------------------------

@add_plone_route("/users", "users", methods=["GET"])
@add_plone_route("/users/<string:username>", "users", methods=["GET"])
def get(context, request, username=None):
    """ Plone users route
    """

    items = []

    if ploneapi.user.is_anonymous():
        raise RuntimeError("Not allowed for anonymous users")

    # list all users if no username was given
    if username is None:
        users = ploneapi.user.get_users()

        for user in users:
            items.append(get_user_info(user.getId()))

    # special user 'current' which retrieves the current user infos
    elif username == "current":
        items.append(get_user_info(short=False))

    # we have a username, go and get the infos for it
    else:
        info = get_user_info(username, short=False)
        items.append(info)

    return {
        "url":   url_for("users"),
        "count": len(items),
        "items": items
    }


@add_plone_route("/auth", "auth", methods=["GET"])
def auth(context, request):
    """ Authenticate
    """

    if ploneapi.user.is_anonymous():
        request.response.setStatus(401)
        request.response.setHeader('WWW-Authenticate', 'basic realm="JSONAPI AUTH"', 1)
    return {}

# vim: set ft=python ts=4 sw=4 expandtab :



# -*- coding: utf-8 -*-

__author__ = 'Ramon Bartl <ramon.bartl@nexiles.com>'
__docformat__ = 'plaintext'

import types


def fail(error):
    """ Raises a RuntimeError with the given error Message

        >>> fail("This failed badly")
        Traceback (most recent call last):
        ...
        RuntimeError: This failed badly
    """
    raise RuntimeError(error)


def truthy(thing):
    """ checks if a value is True or not None

        >>> truthy(0)
        True
        >>> truthy({})
        True
        >>> truthy([])
        True
        >>> truthy(None)
        False
        >>> truthy(False)
        False
    """
    if thing is False or thing is None:
        return False
    return True

def falsy(thing):
    """ checks if a value is False or None

        >>> falsy(0)
        False
        >>> falsy({})
        False
        >>> falsy([])
        False
        >>> falsy(None)
        True
        >>> falsy(False)
        True
    """
    return not truthy(thing)

def is_string(thing):
    """ checks if an object is a string/unicode type

        >>> is_string("")
        True
        >>> is_string(u"")
        True
        >>> is_string(str())
        True
        >>> is_string(unicode())
        True
        >>> is_string(1)
        False
    """
    return type(thing) in types.StringTypes

def is_list(thing):
    """ checks if an object is a list type

        >>> is_list([])
        True
        >>> is_list(list())
        True
        >>> is_list("[]")
        False
        >>> is_list({})
        False
    """
    return type(thing) is types.ListType

def is_tuple(thing):
    """ checks if an object is a tuple type

        >>> is_tuple(())
        True
        >>> is_tuple(tuple())
        True
        >>> is_tuple("()")
        False
        >>> is_tuple([])
        False
    """
    return type(thing) is types.TupleType

def is_dict(thing):
    """ checks if an object is a dictionary type

        >>> is_dict({})
        True
        >>> is_dict(dict())
        True
        >>> is_dict("{}")
        False
        >>> is_dict([])
        False
    """
    return type(thing) is types.DictType

def is_digit(thing):
    """ checks if an object is a digit

        >>> is_digit(1)
        True
        >>> is_digit("1")
        True
        >>> is_digit("a")
        False
        >>> is_digit([])
        False
    """
    return str(thing).isdigit()

def to_int(thing):
    """ coverts an object to int

        >>> to_int("0")
        0
        >>> to_int(1)
        1
        >>> to_int("1")
        1
        >>> to_int("a")

    """
    if is_digit(thing): return int(thing)
    return None

def to_string(thing):
    """ coverts an object to string

        >>> to_string(1)
        '1'
        >>> to_string([])
        '[]'
        >>> to_string(u"a")
        'a'
    """
    return str(thing) or None

def to_list(thing):
    """ converts an object to a list

        >>> to_list(1)
        [1]

        >>> to_list([1,2,3])
        [1, 2, 3]

        >>> to_list(("a", "b", "c"))
        ['a', 'b', 'c']

        >>> to_list(dict(a=1, b=2))
        [{'a': 1, 'b': 2}]
    """
    if not (is_list(thing) or is_tuple(thing)):
        return [thing]
    return list(thing)

def convert(value, converter):
    """ Converts a value with a given converter function.

        >>> convert("1", to_int)
        1
        >>> convert("0", to_int)
        0
        >>> convert("a", to_int)

    """
    if not callable(converter): fail("Converter must be a function")
    return converter(value)

def pluck(col, key, default=None):
    """ Extracts a list of values from a collection of dictionaries

        >>> stooges = [{"name": "moe",   "age": 40},
        ...            {"name": "larry", "age": 50},
        ...            {"name": "curly", "age": 60}]
        >>> pluck(stooges, "name")
        ['moe', 'larry', 'curly']

        It only works with collections

        >>> curly = stooges.pop()
        >>> pluck(curly, "age")
        Traceback (most recent call last):
        ...
        RuntimeError: First argument must be a list or tuple
    """
    if not (is_list(col) or is_tuple(col)):
        fail("First argument must be a list or tuple")
    def _block(dct):
        if not is_dict(dct): return []
        return dct.get(key, default)
    return map(_block, col)

def pick(dct, *keys):
    """ Returns a copy of the dictionary filtered to only have values for the
        whitelisted keys (or list of valid keys)

        >>> pick({"name": "moe", "age": 50, "userid": "moe1"}, "name", "age")
        {'age': 50, 'name': 'moe'}

    """
    copy = dict()
    for key in keys:
        if key in dct.keys(): copy[key] = dct[key]
    return copy

def omit(dct, *keys):
    """ Returns a copy of the dictionary filtered to omit the blacklisted keys
        (or list of keys)

        >>> omit({"name": "moe", "age": 50, "userid": "moe1"}, "userid", "age")
        {'name': 'moe'}
    """
    copy = dict()
    for key in dct:
        if key not in keys: copy[key] = dct[key]
    return copy

def rename(dct, mapping):
    """ Rename the keys of a dictionary with the given mapping

        >>> rename({"a": 1, "BBB": 2}, {"a": "AAA"})
        {'AAA': 1, 'BBB': 2}
    """

    def _block(memo, key):
        if key in dct:
            memo[mapping[key]] = dct[key]
            return memo
        else:
            return memo
    return reduce(_block, mapping, omit(dct, *mapping.keys()))

def alias(col, mapping):
    """ Returns a collection of dictionaries with the keys renamed according to
        the mapping

        >>> libraries = [{"isbn": 1, "ed": 1}, {"isbn": 2, "ed": 2}]
        >>> alias(libraries, {"ed": "edition"})
        [{'edition': 1, 'isbn': 1}, {'edition': 2, 'isbn': 2}]

        >>> alias({"a": 1}, {"a": "b"})
        [{'b': 1}]
    """
    if not is_list(col): col = [col]
    def _block(dct):
        return rename(dct, mapping)
    return map(_block, col)

def first(lst, n=None):
    """ get the first element of a list

        >>> lst = [1, 2, 3, 4, 5]
        >>> first(lst)
        1
        >>> first(lst, 3)
        [1, 2, 3]
    """
    if not is_list(lst): return None
    return n is None and lst[0] or lst[0:n]


if __name__ == '__main__':
    import doctest
    doctest.testmod(raise_on_error=False, optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)

# vim: set ft=python ts=4 sw=4 expandtab :
